# rocketlib.py

import sys, asyncore, asynchat, socket, os, os.path
import string, cgi, traceback, random
import zipfile, StringIO, email
try:
    import email.iterators, email.generator, email.utils
except:
    pass
import webbrowser
from pysqlite2 import dbapi2 as sqlite

#################################################################
### Global Variables
 
db = None
cursor = None
zip = None

types = {
    "jpg":          "image/jpeg",
    "jpeg":         "image/jpeg",
    "png":          "image/png",
    "gif":          "image/gif",
    "css":          "text/css",
}

#################################################################
### Utility Functions
 
def getfile(fn):
    return zip.read(fn)

def quote(s):
    s = string.split(s, '"')
    return string.join(s, "&quot;")

def table_start(w = "width=100%"):
    return "<table border=0 cellspacing=0 cellpadding=0 %s>" % w

def redirect(page, time):
    return """
        <script type=text/javascript>
        setTimeout('location = "%s"', %d*1000 )
        </script>
    """ % (page, time)

def gensortcode(s):
    parts = string.split(s, "-")
    if len(parts) != 2:
        return "?%s" % s
    frac = string.split(parts[0], "A")
    if len(frac) == 2:
        impulse = "%04d" % int(frac[1])
        if not frac[0]:
            desig = "A"
        elif frac[0] == "1/2":
            desig = "1"
        elif frac[0] == "1/2":
            desig = "0"
        else:
            desig = "#"
    else:
        desig = parts[0][0]
        impulse = "%04d" % int(parts[0][1:])
    diam = "18"
    if desig > "C":
        diam = "24"
    if parts[1][-1] == "T":
        diam = "13"
        parts[1] = parts[1][:-1]
    try:
        delay = "%03d" % int(parts[1])
    except:
        delay = "%-3s" % parts[1]
    return "%s%s%s%s" % (diam, desig, impulse, delay)

#################################################################
### "Task" Functions
 
def license(args):
    tmpl = getfile("template")
    data = {}

    data["body"] = getfile("license")
    data["title"] = "Software License"

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def index(args):

    tmpl = getfile("template")
    data = {}

    cursor.execute("""
        select rid, make, model, owner from rockets
        order by make, model
    """)

    body = [ "<table class=sortable id=rockettable",
             "    border=0 cellspacing=0 cellpadding=0 width=100%>",
             "<thead>",
             "<tr><td width=15%>Owner</td>",
             "<td width=15%>Make</td><td width=30%>Model</td>",
             "<td width=1 class=unsortable>&nbsp;&nbsp;</td>",
             "<td width=40% class=unsortable>Engines</td></tr>",
             "</thead>",
           ]

    rows = cursor.fetchall()

    for row in rows:
        # get recommendations
        cursor.execute("""
            select recommended.designation
            from recommended left join engines on engines.designation = recommended.designation
            where rid = ?
            order by sortcode, recommended.designation
        """, (row[0],))
        engines = map(lambda x: "<a href='whatneeds?d=%s'>%s</a>" % (x[0], x[0]), cursor.fetchall())
        # format the results
        body.append("<tr valign=top class=underscore>")
        body.append("<td>%s</td>" % row[3])
        body.append("<td>%s</td>" % row[1])
        body.append("<td><a href='showrocket?rid=%d'>" % row[0])
        body.append("%s&nbsp;</a></td><td></td>" % row[2])
        body.append("<td>%s</td></tr>" % string.join(engines, ", "))

    body.append("</table>")

    data["body"] = string.join(body, "\n")
    data["title"] = "List of All Rockets"

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def addengine(args):
    newengine = args["newengine"].strip().upper()
    cursor.execute("""
        insert into engines (designation)
        values (?)
    """, (newengine,))
    db.commit()

    return string.join([
        "HTTP/1.0 302 Found",
        "Location: /inventory\n",
    ], "\n")

def addrocket(args):
    cursor.execute("""
        insert into rockets (make)
        values ("")
    """)
    rid = cursor.lastrowid
    db.commit()
    return string.join([
        "HTTP/1.0 302 Found",
        "Location: /editrocket?rid=%d\n" % int(rid),
    ], "\n")

def editrocket(args):
    tmpl = getfile("template")
    data = {}

    rid = int(args["rid"])

    cursor.execute("""
        select make, model, notes, owner from rockets
        where rid = ?
    """, (rid,))
    rocket = cursor.fetchone()

    cursor.execute("""
        select designation from recommended
        where rid = ?
        order by designation
    """, (rid,))
    recommended = map(lambda x: x[0], cursor.fetchall())

    cursor.execute("""
        select designation from engines
        order by sortcode, designation
    """)
    engines = map(lambda x: x[0], cursor.fetchall())

    title = "Update %s (%s)" % (rocket[1], rocket[0])
    body = [ "<form method=get action=updaterocket>",
             "<input type=hidden name=rid value='%d'>" % rid,
             "<table border=0 cellspacing=2 cellpadding=0>",
             "<tr><td><b>Make&nbsp;</b></td>",
             '<td><input type=text name=make value="%s"></td></tr>' % quote(rocket[0]),
             "<tr><td><b>Model&nbsp;</b></td>",
             '<td><input type=text name=model value="%s"></td></tr>' % quote(rocket[1]),
             "<tr><td><b>Owner&nbsp;</b></td>",
             '<td><input type=text name=owner value="%s"></td></tr>' % quote(rocket[3]),
             "<tr><td><b>Notes&nbsp;</b></td>",
             '<td><input type=text name=notes value="%s"></td></tr>' % quote(rocket[2]),
             "<tr><td><b>Engines&nbsp;</b></td></tr>",
           ]

    for engine in engines:
        body.append("<tr><td>&nbsp;&nbsp;%s&nbsp;&nbsp;</td>" % engine)
        if engine in recommended:
            body.append("<td><input type=checkbox name=e value='%s' checked>" % engine)
        else:
            body.append("<td><input type=checkbox name=e value='%s'>" % engine)
        body.append("</tr>")

    body.append("</table>")
    body.append("<input type=submit value='Save Changes'>")
    body.append("</form>")

    data["body"] = string.join(body, "\n")
    data["title"] = title

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def engineupdate(args):
    d = args["d"].strip()
    q = int(args["q"])
    cursor.execute("""
        update engines set quantity = ? where designation = ?
    """, (q, d))
    db.commit()
    return string.join([
        "HTTP/1.0 200 OK Response Follows",
        "Content-type: text/plain\n",
        "OK."
    ], "\n")

def flyable(args):
    tmpl = getfile("template")
    data = {}

    cursor.execute("""
        select rockets.rid, make, model, rockets.notes,
            recommended.designation,
            recommended.notes as rnotes, quantity, sortcode, owner
        from rockets inner join recommended on rockets.rid = recommended.rid
            inner join engines on engines.designation = recommended.designation
        where quantity > 0
        order by make || model, sortcode
    """)

    current = ("", "")

    row = cursor.fetchone()

    body = [ "<center>\n<table border=0 class=sortable id=rockettable",
             "    cellspacing=0 cellpadding=0 width=100%>",
             "<thead>",
             "<tr><td width=15%>Owner</td>",
             "<td width=15%>Make</td><td width=30%>Model</td>",
             "<td width=1 class=unsortable>&nbsp;&nbsp;</td>",
             "<td width=40% class=unsortable><b>Engines</b></td></tr>",
             "</thead>",
        ]

    engines = []

    while row:
        c = (row[1], row[2])
        if c != current:
            if engines:
                engines = map(lambda x: "<a href='whatneeds?d=%s'>%s</a>" % (x, x), engines)
                body.append("<td>%s</td></tr>" % string.join(engines, ", "))
                engines = []
            body.append("<tr valign=top>")
            body.append("<td>%s</td>" % row[8])
            body.append("<td>%s</td>" % row[1])
            body.append("<td><a href='showrocket?rid=%d'>" % row[0])
            body.append("%s</a></td><td></td>" % row[2])
            current = c
        engines.append(row[4])
        row = cursor.fetchone()

    if engines:
        body.append("<td>%s</td></tr></table>" % string.join(engines, ", "))

    data["body"] = string.join(body, "\n")
    data["title"] = "What Can We Fly Today?"

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def inventory(args):
    tmpl = getfile("template")
    data = {}

    cursor.execute("""
        select designation, quantity
        from engines
        order by sortcode, designation
    """)

    body = [ "<script src='js/engineupdate.js' type='text/javascript'></script>",
             "<center><table border=0 cellspacing=0 cellpadding=0 width=50%>",
             "<tr><td><b>Engine</b></td><td width=1>&nbsp;&nbsp;</td>",
             "<td align=right><b>Quantity</b></td></tr>",
           ]

    bgcolor = [ "#cccccc", "white" ]

    row = cursor.fetchone()
    idx = 0

    while row:
        idx += 1
        body.append("<tr valign=top bgcolor=%s id='tr%d'>" % (bgcolor[0], idx))
        body.append("<td><a href='whatneeds?d=%s'>%s</a></td>" % (row[0], row[0]))
        body.append("<td></td><td align=right><input type=text value='%d'" % row[1])
        body.append("size=5 maxlength=5 id=q%d" % idx)
        body.append("style='background-color: %s; text-align: right;'" % bgcolor[0])
        body.append("""onchange='engineupdate("%s", %d)'>""" % (row[0], idx))
        body.append("<input type=hidden value='%s' name=designation>" % row[0])
        body.append("""<input type=button value='X' title='Delete' onclick='location = "deleteengine?d=%s"'>""" % row[0])
        body.append("</td></tr>")
        bgcolor = [ bgcolor[1], bgcolor[0] ]
        row = cursor.fetchone()

    body.append("<tr valign=top bgcolor=white>")
    body.append("<td colspan=3>")
    body.append("<form method=get action=addengine>")
    body.append("<table border=0 cellspacing=0 cellpadding=0 width=100%% bgcolor='%s'>" % bgcolor[0])
    body.append("<tr valign=top>")
    body.append("<td><input type=text name=newengine></td>")
    body.append("<td align=right><input type=submit value='Add'></td>")
    body.append("</tr></table></form>")
    body.append("</td></tr>")

    body.append("</table></center>")

    data["body"] = string.join(body, "\n")
    data["title"] = "Engine Inventory"

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")


def showrocket(args):
    tmpl = getfile("template")
    data = {}

    rid = int(args["rid"])

    cursor.execute("""
        select rockets.rid, make, model, rockets.notes,
            recommended.designation,
            recommended.notes as rnotes, quantity, owner
        from rockets left join recommended on rockets.rid = recommended.rid
            left join engines on engines.designation = recommended.designation
        where rockets.rid = ?
        order by engines.sortcode, recommended.designation
    """, (rid,))

    row = cursor.fetchone()

    title = "%s (%s)" % (row[2], row[1])
    body = [ "<table border=0 cellspacing=2 cellpadding=0 width=100%>",
            "<tr valign=top><td width=80%>",
            title,
            "<br>",
        ]

    if row[7].strip():
        body.append("Owner: <b>%s</b><br>" % row[7])

    if row[3].strip():
        body.append("%s<br>" % row[3])

    engines = []

    while row:
        if row[6] > 0:
            engines.append("<a href='whatneeds?d=%s'><b>%s [%d]</b></a>" % (row[4], row[4], row[6]))
        else:
            engines.append("<a href='whatneeds?d=%s' style='color: gray;'>%s</a>" % (row[4], row[4]))
        row = cursor.fetchone()

    body.append(string.join(engines, ", "))

    body.append("</td><td width=20% align=right>")
    body.append("""<input type=button value='Update' style='width: 8em;' onclick='location = "editrocket?rid=%d"'>""" % rid)
    body.append("<br>")
    body.append("""<input type=button value='Delete' style='width: 8em;' onclick='location = "askdeleterocket?rid=%d"'>""" % rid)
    body.append("<br>")
    body.append("</td></tr></table>")

    data["body"] = string.join(body, "\n")
    data["title"] = title

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")


def updaterocket(args):
    rid = int(args["rid"])
    cursor.execute("""
        update rockets set
            make = ?, model = ?, notes = ?, owner = ?
        where rid = ?
    """, (args.get("make", ""), args.get("model", ""), args.get("notes", ""), args.get("owner", ""), rid))
    try:
        recommended = args["e"]
    except:
        recommended = []
    if type(recommended) is type(""):
        recommended = [ recommended ]
    # clear existing recommendations
    cursor.execute("""
        delete from recommended
        where rid = ?
    """, (rid,))
    # store new recommendations
    for engine in recommended:
        cursor.execute("""
            insert into recommended (rid, designation)
            values (?, ?)
        """, (rid, engine))
    db.commit()
    return string.join([
        "HTTP/1.0 302 Found",
        "Location: /showrocket?rid=%d\n" % int(rid),
    ], "\n")

def whatneeds(args):
    tmpl = getfile("template")
    data = {}

    d = args["d"].strip()

    cursor.execute("""
        select rockets.rid, make, model, rockets.notes
        from rockets inner join recommended on rockets.rid = recommended.rid
        where recommended.designation = ?
        order by model, make
    """, (d,))

    current = ("", "")

    body = [ "<table border=0 cellspacing=0 cellpadding=0 width=100%>",
             "<tr valign=top><td width=80%>",
           ]

    rows = cursor.fetchall()

    for row in rows:
        body.append("<a href='showrocket?rid=%d'>" % row[0])
        rocket = ("%s <b>%s</b>" % (row[1], row[2])).strip()
        if rocket == "<b></b>":
            rocket = "(Unnamed)"
        body.append("%s<br></a>" % rocket)

    body.append("""
    </td><td width=20% align=right>
    <input type=button value='Add Rocket' onclick='location = "addrocket"'>
    </td></tr>
    </table>
    """)

    data["body"] = string.join(body, "\n")
    data["title"] = "What Needs %s?" % d

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def cantfly(args):
    tmpl = getfile("template")
    data = {}

    cursor.execute("""
        select rockets.rid, make, model, sum(quantity) as qsum, owner
        from rockets inner join recommended on rockets.rid = recommended.rid
            inner join engines on engines.designation = recommended.designation
        group by rockets.rid, make, model
        order by make, model
    """)

    rows = cursor.fetchall()

    body = [ "<table class=sortable id=rockettable",
             "    border=0 cellspacing=0 cellpadding=0 width=100%>",
             "<thead>",
             "<tr><td width=15%>Owner</td>",
             "<td width=15%>Make</td><td width=30%>Model</td>",
             "<td width=1 class=unsortable>&nbsp;&nbsp;</td>",
             "<td width=40% class=unsortable>Engines</td></tr>",
             "</thead>",
           ]

    for row in rows:
        if row[3] == 0:
            # get recommendations
            cursor.execute("""
                select designation
                from recommended
                where rid = ?
                order by designation
            """, (row[0],))
            engines = map(lambda x: "<a href='whatneeds?d=%s'>%s</a>" % (x[0], x[0]), cursor.fetchall())
            # format the results
            body.append("<tr valign=top>")
            body.append("<td>%s</td>" % row[4])
            body.append("<td><a href='showrocket?rid=%d'>" % row[0])
            body.append("%s</a></td>" % row[1])
            body.append("<td><a href='showrocket?rid=%d'>" % row[0])
            body.append("%s</a></td><td></td>" % row[2])
            body.append("<td>%s</td></tr>" % string.join(engines, ", "))

    body.append("</table>")

    data["body"] = string.join(body, "\n")
    data["title"] = "What Can We NOT Fly Today?"

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def deleterocket(args):
    tmpl = getfile("template")
    data = {}

    rid = int(args["rid"])

    cursor.execute("""
        select make, model, notes
        from rockets
        where rid = ?
    """, (rid,))

    row = cursor.fetchone()

    cursor.execute("""
        delete from rockets
        where rid = ?
    """, (rid,))

    cursor.execute("""
        delete from recommended
        where rid = ?
    """, (rid,))

    db.commit()

    title = "%s (%s)" % (row[1], row[0])
    title = title.strip()
    if title == "()":
        title = "Unknown (ID = %d)" % rid

    body = [ "<center>\n&nbsp;\n<p>",
             "<b class=title>%s Deleted.</b>" % title,
             "</center>",
             "<p>&nbsp;",
             redirect("index", 3),
           ]

    data["body"] = string.join(body, "\n")
    data["title"] = "%s Deleted." % title

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")


def askdeleterocket(args):
    tmpl = getfile("template")
    data = {}

    rid = int(args["rid"])

    cursor.execute("""
        select make, model, notes
        from rockets
        where rid = ?
    """, (rid,))

    row = cursor.fetchone()

    title = "%s (%s)" % (row[1], row[0])
    title = title.strip()
    if title == "()":
        title = "Unknown (ID = %d)" % rid

    body = [ "<center>\n&nbsp;\n<p>",
             "<b class=title>Delete %s?</b>\n<p>" % title,
             table_start(w = ""),
             "<tr><td>",
             """<input type=button value='Yes' style='width: 8em;' onclick='location = "deleterocket?rid=%d"'>""" % rid,
             "</td><td>",
             """<input type=button value='No' style='width: 8em;' onclick='location = "showrocket?rid=%d"'>""" % rid,
             "</td></tr></table>",
             "</center>",
             "<p>&nbsp;",
           ]

    data["body"] = string.join(body, "\n")
    data["title"] = "Delete %s?" % title

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def deleteengine(args):
    tmpl = getfile("template")
    data = {}

    d = args["d"].strip()

    cursor.execute("""
        select quantity from engines
        where designation = ?
    """, (d,))

    engine = cursor.fetchone()

    # should be one.
    if not engine:
        # deal with missing record.
        body = [ "<center>\n&nbsp;\n<p>",
                 "<b class=title>%s: No Such Engine.</b>" % d,
                 "</center>",
                 "<p>&nbsp;",
               ]
        data["title"] = "Error"
    else:
        # we have an engine record.  check for recommendations.

        cursor.execute("""
            select rockets.rid, make, model
            from rockets inner join recommended on rockets.rid = recommended.rid
            inner join engines on engines.designation = recommended.designation
            where recommended.designation = ?
            order by model, make
        """, (d,))

        rockets = cursor.fetchall()

        if engine[0] > 0 or len(rockets) > 0:
            # let the user know.
            body = [ "<b class=subtitle>Warning: Deleting %s</b>\n<p>" % d,
                   ]

            if engine[0] > 0:
                body.append("There are %d %s engines in inventory.\n<p>" % (engine[0], d))

            if len(rockets) > 0:
                body.append("The following rockets recommend this engine:<br>")
                for rocket in rockets:
                    txt = ("%s <b>%s</b>" % (rocket[1], rocket[2])).strip()
                    if txt == "<b></b>":
                        txt = "(Unnamed)"
                    body.append("<a href='showrocket?rid=%d'>" % rocket[0])
                    body.append("&nbsp;&nbsp;%s</a><br>" % txt)

            body.append("<p>Delete this engine anyway?\n<p align=center>")
            body.append(table_start(w = ""))
            body.append("<tr><td>")
            body.append("""<input type=button value='Yes' style='width: 8em;' onclick='location = "forcedeleteengine?d=%s"'>""" % d)
            body.append("</td><td>")
            body.append("""<input type=button value='No' style='width: 8em;' onclick='location = "inventory"'>""")
            body.append("</td></tr></table>\n<p>&nbsp;")

            data["title"] = "Warning: Deleting %s" % d

        else:
            # no reason not to kill the record now...
            cursor.execute("""
                delete from engines
                where designation = ?
            """, (d,))
            db.commit()
            body = [ "<center>\n&nbsp;\n<p>",
                     "<b class=title>%s Deleted.</b>" % d,
                     "</center>",
                     "<p>&nbsp;",
                     redirect("inventory", 2),
                   ]
            data["title"] = "%s Deleted." % d

    data["body"] = string.join(body, "\n")

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def forcedeleteengine(args):

    tmpl = getfile("template")
    data = {}

    d = args["d"].strip()

    cursor.execute("""
        delete from engines
        where designation = ?
    """, (d,))

    cursor.execute("""
        delete from recommended
        where designation = ?
    """, (d,))

    db.commit()

    body = [ "<center>\n&nbsp;\n<p>",
             "<b class=title>%s Deleted.</b>" % d,
             "</center>",
             "<p>&nbsp;",
             redirect("inventory", 2),
           ]

    data["title"] = "%s Deleted." % d
    data["body"] = string.join(body, "\n")

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         tmpl % data
    ], "\n")

def shutdown(args):

    server.close()

    return string.join([
         "HTTP/1.0 200 OK Response Follows",
         "Content-type: text/html\n",
         "<html>\n<head>\n<title>Server Shutdown Complete</title>",
         "</head>\n<body bgcolor=white>",
         "<p>&nbsp;",
         "<p>&nbsp;",
         "<center>",
         "<table border=0 cellspacing=2 cellpadding=5 bgcolor=black>",
         "<tr><td bgcolor=white align=center>",
         "<b style='font-size: 150%;'>&nbsp;Server&nbsp;Shutdown&nbsp;Complete&nbsp;</b>",
         "<p>\n(Please close this browser window.)",
         "</td></tr></table>",
         "</center>",
    ], "\n")

tasks = {
    "addengine":         addengine,
    "addrocket":         addrocket,
    "cantfly":           cantfly,
    "deleterocket":      deleterocket,
    "askdeleterocket":   askdeleterocket,
    "deleteengine":      deleteengine,
    "forcedeleteengine": forcedeleteengine,
    "editrocket":        editrocket,
    "engineupdate":      engineupdate,
    "flyable":           flyable,
    "index":             index,
    "inventory":         inventory,
    "license":           license,
    "showrocket":        showrocket,
    "updaterocket":      updaterocket,
    "whatneeds":         whatneeds,
    "quit":              shutdown,
}

#################################################################
### Main Task Dispatcher

def getpage(path, method, cgidata):

    path_lst = string.split(path, "?", maxsplit=1)
    path = path_lst[0]
    if len(path_lst) > 1:
        targs = cgi.parse_qsl(path_lst[1])
    else:
        targs = []
    args = {}
    for v, d in targs:
        if args.has_key(v):
            if type(args[v]) is type(""):
                args[v] = [ args[v], d ]
            else:
                args[v].append(d)
        else:
            args[v] = d

    if tasks.has_key(path):
        return tasks[path](args)

    # not an internal task, must be a file.

    xpath = string.split(path, "/")
    npath = []
    for p in xpath:
        if p != "" and p != "..":
            npath.append(p)
        else:
            npath = []

    path = string.join(npath, "/")

    k = string.split(path, ".")[-1]
    try:
        typ = types[k]
    except:
        typ = "application/octet-stream"

    try:
        rc = [
            "HTTP/1.0 200 OK Response Follows",
            "Content-type: %s" % typ,
            "",
            getfile(path),
        ]

    except :
        rc = [
            "HTTP/1.0 200 OK Message Follows",
            "Content-type: text/plain"
            "",
            traceback.format_exc(),
        ]

    return string.join(rc, "\n")

#################################################################
### Internal Web Server

class mainserver(asyncore.dispatcher):

    def __init__(self, ip, port):
        self.ip= ip
        self.port = port
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind((ip, port))
        self.listen(5)

    def readable(self):
        return self.accepting

    def handle_accept(self):
        try:
            conn, addr = self.accept()
        except socket.error: # rare Linux error
            print "Socket error on server accept()"
            return
        except TypeError: # rare FreeBSD3 error
            print "EWOULDBLOCK exception on server accept()"
            return
        handler(conn, addr)

class handler(asynchat.async_chat):

    def __init__(self, conn, addr):
        asynchat.async_chat.__init__(self, conn=conn)
        self.addr = addr
        self.ibuffer = []
        self.obuffer = ""
        self.set_terminator("\n")
        self.header_lines = []
        self.reading_headers = True
        self.request_line = None
        self.cgi_data = None
        self.headers = None

    def collect_incoming_data(self, data):
        self.ibuffer.append(data.replace('\r', ''))

    def found_terminator(self):
        if self.reading_headers:
            txt = "".join(self.ibuffer).strip()
            if self.request_line is None:
                self.request_line = txt
            elif txt == "":
                self.reading_headers = False
                self.header_lines.append("")
                self.headers = email.message_from_string("\n".join(self.header_lines))
                self.method, self.path, self.protocol = self.request_line.split()
                if self.method.upper() == "POST":
                    clen = self.headers["content-length"]
                    self.set_terminator(int(clen))
                else:
                    self.handling = True
                    self.set_terminator(None)
                    self.handle_request()
            else:
                self.header_lines.append(txt)
        elif not self.handling:
            self.set_terminator(None)
            self.cgi_data = "".join(self.ibuffer)
            self.handling = True
            self.handle_request()
        self.ibuffer = []

    def handle_request(self):
        self.path = self.path[1:]
        if not self.path:
            self.path = "index"
        try:
            self.obuffer = getpage(self.path, self.method, self.cgi_data)
        except:
            self.obuffer = string.join([
                "HTTP/1.0 500 Server Error",
                "Content-type: text/plain",
                "",
                "An Error Has Occured:\n",
                traceback.format_exc(),
            ], "\n")

    def writable(self):
        return len(self.obuffer) > 0

    def handle_write(self):
        sent = self.send(self.obuffer)
        self.obuffer = self.obuffer[sent:]
        if len(self.obuffer) == 0:
            self.close()

#################################################################
### Main Function
 
def run():
    global server, db, cursor, zip

    print "New Century Rocketry"
    print "Rocket Database Server\n"

    port = 8080 + random.randint(1, 9)

    server = mainserver('', port)

    zip = zipfile.ZipFile("rockets.zip")

    if not os.path.exists("rocketry.db"):
        print "First Run, Initializing Database\n"
        rdb = getfile("rocketry.db")
        fp = open("rocketry.db", "wb")
        fp.write(rdb)
        fp.close()

    db = sqlite.connect("rocketry.db")
    cursor = db.cursor()

    try:
        cursor.execute('''select value from globalv where key = "version"''')
        row = cursor.fetchall()[0]
        dbversion = int(row[0])
    except sqlite.OperationalError:
        dbversion = 0

    if dbversion < 1:
        print "Upgrade Original Database to Version 1 Database."

        db.close()

        os.rename("rocketry.db", "rocketry-old.db")
        rdb = getfile("rocketry.db")
        fp = open("rocketry.db", "wb")
        fp.write(rdb)
        fp.close()

        db = sqlite.connect("rocketry.db")
        cursor = db.cursor()

        dbold = sqlite.connect("rocketry-old.db")
        oldcursor = dbold.cursor()

        oldcursor.execute("""
            select designation, quantity
            from engines
        """)
        rows = oldcursor.fetchall()
        for row in rows:
            sortcode = gensortcode(row[0])
            cursor.execute("""
                insert into engines
                    (designation, sortcode, quantity)
                values
                    (?, ?, ?)
            """, (row[0], sortcode, row[1]))

        oldcursor.execute("""
            select rid, designation, notes
            from recommended
        """)
        rows = oldcursor.fetchall()
        for row in rows:
            cursor.execute("""
                insert into recommended
                    (rid, designation, notes)
                values
                    (?, ?, ?)
            """, (row[0], row[1], row[2]))
    
        oldcursor.execute("""
            select rid, make, model, notes
            from rockets
        """)
        rows = oldcursor.fetchall()
        for row in rows:
            cursor.execute("""
                insert into rockets
                    (rid, make, model, notes)
                values
                    (?, ?, ?, ?)
            """, (row[0], row[1], row[2], row[3]))

        dbold.close()
        db.commit()

        dbversion = 1

    if dbversion < 2:
        print "Upgrade Version 1 Database to Version 2 Database."

        db.close()

        try:
            os.remove("rocketry-old.db")
        except:
            pass

        os.rename("rocketry.db", "rocketry-old.db")
        rdb = getfile("rocketry.db")
        fp = open("rocketry.db", "wb")
        fp.write(rdb)
        fp.close()

        db = sqlite.connect("rocketry.db")
        cursor = db.cursor()

        dbold = sqlite.connect("rocketry-old.db")
        oldcursor = dbold.cursor()

        oldcursor.execute("""
            select designation, sortcode, quantity
            from engines
        """)
        rows = oldcursor.fetchall()
        for row in rows:
            cursor.execute("""
                insert into engines
                    (designation, sortcode, quantity)
                values
                    (?, ?, ?)
            """, (row[0], row[1], row[2]))

        oldcursor.execute("""
            select rid, designation, notes
            from recommended
        """)
        rows = oldcursor.fetchall()
        for row in rows:
            cursor.execute("""
                insert into recommended
                    (rid, designation, notes)
                values
                    (?, ?, ?)
            """, (row[0], row[1], row[2]))
    
        oldcursor.execute("""
            select rid, make, model, notes
            from rockets
        """)
        rows = oldcursor.fetchall()
        for row in rows:
            cursor.execute("""
                insert into rockets
                    (rid, make, model, notes)
                values
                    (?, ?, ?, ?)
            """, (row[0], row[1], row[2], row[3]))

        
        dbold.close()
        db.commit()

        dbversion = 2

    print "Server Initialized, Port", port

    print "\n(If you forget to click Exit in the web interface,"
    print "and need to close this window, press CTRL-BREAK.)"

    webbrowser.open("http://127.0.0.1:%d/" % port)

    try:
        asyncore.loop(4.0) # Frequent checks for interrupt
    except KeyboardInterrupt:
        print "Server Terminated."


# end of file.
